#include "cnetheader.h"

/*  The cnet network simulator (v1.7.7)
    Copyright (C) 1992-2002, Chris McDonald

    Chris McDonald, chris@cs.uwa.edu.au
    Department of Computer Science & Software Engineering
    The University of Western Australia,
    Crawley, Western Australia, 6009
    PH: +61 8 9380 2533, FAX: +61 8 9380 1089.

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*  ------------ Nothing in here for USE_ASCII compilation ------------ */

#if	defined(USE_TCLTK)

#define	N_DFS		32
static	DRAWFRAME	*dfs[N_DFS]	= {0};
static	int		ndfs		= 0;

void new_drawframe(int len, char *frame, long delay, long Tprop,
		   int is_lost, int is_corrupt)
{
    extern	void schedule_event(CnetEvent ne, int node, long rel_ms,
				    CnetTimer timer, CnetData data);
    DRAWFRAME	*df;

    df			= NEW(DRAWFRAME);
    memset(df, 0, sizeof(DRAWFRAME));
    df->started		= TIMENOW_in_MSEC + delay;
    df->arrives		= df->started + Tprop;
    df->inflight	= df->arrives - df->started;
    df->when_lost	= is_lost    ? (df->started + Tprop/2) : MAXINT;
    df->when_corrupt	= is_corrupt ? (df->started + Tprop/2) : MAXINT;

    df->cdf.len		= len;
    df->cdf.frame	= malloc(len);
    memcpy(df->cdf.frame, frame, len);

/*  USER-CODE WILL ADD THE COLOURS AND PIXEL LENGTHS REQUIRED */
    schedule_event(EV_DRAWFRAME, THISNODE, delay, NULLTIMER, (CnetData)df);
}

void free_drawframe(DRAWFRAME *df)
{
    free(df->cdf.frame);
    free(df);
}

/*  CALLED AFTER USER-CODE HAS ADDED THE COLOURS AND PIXEL LENGTHS REQUIRED */
void add_drawframe(DRAWFRAME *df)
{
    char	buf[BUFSIZ], *s;
    int		n, nf, v;

    free(df->cdf.frame);
    df->cdf.len	= 0;

/*  ENSURE THAT AT LEAST ONE VALID COLOUR AND LENGTH WAS PROVIDED */
    for(nf=0 ; nf<CN_NCOLOURS ; ++nf)
	if(df->cdf.colour[nf] <= 0 || df->cdf.pixels[nf] <= 0)
	    break;
    if(nf == 0) {
	free(df);
	return;
    }

/*  ENSURE THAT WE HAVE A FREE SLOT TO REMEMBER THIS DRAWFRAME */
    for(n=0 ; n<N_DFS ; ++n)
	if(dfs[n] == (DRAWFRAME *)NULL)
	    break;
    if(n == N_DFS) {
	free(df);
	return;
    }

    dfs[n]	= df;
    ++ndfs;

    df->srcnode	= THISNODE;

/*  BUILD THE TCL/TK COMMAND IN buf */
    s	= buf;
    sprintf(buf, "DrawFrame %s %d %d %d df%d ",
		argv0, nf, NP[THISNODE].nattr.x, THISNODE == 0 ? 10 : 25, n);
    while(*s) ++s;
    *s++	= '{';
    for(n=0 ; n<nf ; ++n) {
	sprintf(s, "%d ", ((df->cdf.colour[n]-1) % CN_NCOLOURS)+1);
	while(*s) ++s;
    }
    *s++	= '}';
    *s++	= ' ';
    *s++	= '{';
    for(n=0 ; n<nf ; ++n) {
	sprintf(s, "%d ", df->cdf.pixels[n] % 101);
	while(*s) ++s;
    }
    *s++	= '}';
    sprintf(s, " %d", THISNODE == 0 ? -1 : 1);
    v		= vflag;
    vflag	= FALSE;
    TCLTK(buf);
    vflag	= v;
}

void lose_all_drawframes(void)
{
    int	n, v = vflag;

    vflag	= FALSE;
    for(n=0 ; n<N_DFS ; ++n)
	if(dfs[n]) {
	    TCLTK(".%s.df itemconfigure df%d -fill white", argv0, n);
	    dfs[n]->when_lost	= MAXINT;
	}
    vflag	= v;
}


/*  RE-COLOUR, DELETE, AND MOVE ALL OF THE FRAMES CURRENTLY IN TRANSIT */

void move_drawframes()
{
    static long	zap_time = MAXINT;
    DRAWFRAME	*df;
    int		n, totalx, v;

/*  ANYTHING TO DRAW? */
    if(ndfs == 0)
	return;

    v		= vflag;
    vflag	= FALSE;

    totalx	= NP[1].nattr.x - NP[0].nattr.x;
    for(n=0 ; n<N_DFS ; ++n) {
	if(dfs[n] == (DRAWFRAME *)NULL)
	    continue;
	df	= dfs[n];

/*  STILL MOVING */
	if(TIMENOW_in_MSEC < df->arrives) {
	    int		x	= totalx *
		   (((float)(TIMENOW_in_MSEC - df->started)) / df->inflight);
	    int		dx	= x - df->x;

/*  IF FRAME GETS LOST, CHANGE ITS COLOUR TO WHITE */
	    if(df->when_lost < TIMENOW_in_MSEC) {
		TCLTK(".%s.df itemconfigure df%d -fill white", argv0, n);
		df->when_lost		= MAXINT;
	    }
/*  IF FRAME GETS CORRUPTED, ZAP IT, AND CHANGE ITS COLOUR TO DARK-GREY */
	    else if(df->when_corrupt < TIMENOW_in_MSEC) {
		TCLTK(
	    ".%s.df create image %d %d -image im_zap -anchor w -tags dfzap",
			argv0, (totalx + NP[0].nattr.x)/2, 22);
		zap_time	= TIMENOW_in_MSEC+400;

		TCLTK(".%s.df itemconfigure df%d -fill grey50", argv0, n);
		df->when_corrupt	= MAXINT;
	    }
	    else if(zap_time < TIMENOW_in_MSEC) {
		TCLTK(".%s.df delete dfzap", argv0);
		zap_time	= MAXINT;
	    }

	    df->x	+= dx;
	    TCLTK(".%s.df move df%d %d 0", argv0,n,(df->srcnode==0) ? dx : -dx);
	}
/*  REACHED THE END OF ITS TRAVEL */
	else {
	    TCLTK(".%s.df delete df%d", argv0, n);
	    free(df);
	    dfs[n]	= (DRAWFRAME *)NULL;
	    --ndfs;
	}
    }
    vflag	= v;
}

#endif		/* defined(USE_TCLTK) */
