#include "cnetheader.h"

/*  The cnet network simulator (v1.7.7)
    Copyright (C) 1992-2002, Chris McDonald

    Chris McDonald, chris@cs.uwa.edu.au
    Department of Computer Science & Software Engineering
    The University of Western Australia,
    Crawley, Western Australia, 6009
    PH: +61 8 9380 2533, FAX: +61 8 9380 1089.

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

int CNET_set_debug_string(CnetEvent ev, char *str)
{
    if(str == (char *)NULL)
	str = "";
    if((int)ev<(int)EV_DEBUG1 || (int)ev >= ((int)EV_DEBUG1 + N_CNET_DEBUGS)) {
	if(gattr.trace_events)
	    TRACE("\tCNET_set_debug_string(ev=%d,\"%s\") = -1 %s\n",
				(int)ev,str,cnet_errname[(int)ER_BADEVENT]);
	cnet_errno	= ER_BADEVENT;
	return (-1);
    }
    if(gattr.trace_events)
	TRACE("\tCNET_set_debug_string(%s,\"%s\") = 0\n",
				cnet_evname[(int)ev],str);

#if	defined(USE_TCLTK)

    if(Wflag) {
#define	DEBUG_STRING_WIDTH	12
	int	d = (int)ev-(int)EV_DEBUG1;

	if(NP[THISNODE].debug_str[d])
	    free(NP[THISNODE].debug_str[d]);

	if(*str == '\0')
	    NP[THISNODE].debug_str[d]	= strdup("");
	else {
	    strncpy(chararray,str,DEBUG_STRING_WIDTH);
	    chararray[DEBUG_STRING_WIDTH]	= '\0';
	    NP[THISNODE].debug_str[d]	= strdup(chararray);
	}

	str = NP[THISNODE].debug_str[d];
	TCLTK("SetDebugString .node%d.d.debug%d \"%s\"",
		THISNODE, d, (str == (char *)NULL || *str == '\0') ? " " : str);
#undef	DEBUG_STRING_WIDTH
    }

#endif	/* defined(USE_TCLTK) */

    return 0;
}

/* ------------------ ONLY CODE FOR USE_TCLTK --------------------- */

#if	defined(USE_TCLTK)

#define	N_NODE_CHOICES		 4

static struct {
    char	*title;
    char	*keyword;
    char	*labels[N_NODE_CHOICES];
    int		 panel_value;
    int		 values[N_NODE_CHOICES];
    int		   orig[N_NODE_CHOICES];
} node_defaults[] = {

    { "Message Rate :",		"messagerate",	{ NULL, NULL, NULL, NULL },
    				0,		{   0, 0, 0, 0 },
    {  200, 10, 500, 5000 }},	/* in ms */

    { "Min. Message Size :",	"minmessagesize", { NULL, NULL, NULL, NULL },
    				0,		{   0, 0, 0, 0 },
    {  100, 32, 256, K }},

    { "Max. Message Size :",	"maxmessagesize", { NULL, NULL, NULL, NULL },
    				0,		{   0, 0, 0, 0 },
    {  K, 2*K, 4*K, MAX_MESSAGE_SIZE  }}
};

#define	N_NODE_DEFAULTS		(sizeof(node_defaults)/sizeof(node_defaults[0]))

#define	RATE_CHOICE		0
#define	MINSIZE_CHOICE		1
#define	MAXSIZE_CHOICE		2

typedef	struct	{
    int		n;
    int		r;
    int		c;
    int		value;
} NODECHOICE;


/* ---------------------------------------------------------------------- */


static int node_debug_button(ClientData data, Tcl_Interp *interp,
				int argc, char *argv[])
{
    extern void	single_event(int, CnetEvent, CnetTimer);
    extern void schedule_event(CnetEvent, int, long, CnetTimer, CnetData);

    CnetEvent	ev;
    int		d, n;

    if(vflag) {
	for(n=0 ; n<argc-1 ; ++n)
	    fprintf(stderr,"%s ", argv[n]);
	fprintf(stderr,"%s\n", argv[argc-1]);
    }
    if(argc != 3) {
	interp->result	= "wrong # args";
	return TCL_ERROR;
    }
    n	= atoi(argv[1]);
    d	= atoi(argv[2]);
    if(n < 0 || n >= _NNODES || d < 0 || d >= N_CNET_DEBUGS) {
	interp->result	= "invalid node or debug #";
	return TCL_ERROR;
    }

    ev	= (CnetEvent)((int)EV_DEBUG1+d);
    if(dflag)
	fprintf(stderr,"%s.%s\n", NP[n].nodename, cnet_evname[(int)ev]);

    if(cnet_state == STATE_PAUSED)
	single_event(n, ev, NULLTIMER);
    else
	schedule_event(ev, n, 1L, NULLTIMER, NP[n].data[(int)ev]);

    return TCL_OK;
}



void fmt_choice(int i, int div, char *shortfmt, char *longfmt)
{
    if((i%div) == 0)	sprintf(chararray,shortfmt,i/div);
    else		sprintf(chararray,longfmt, i);
}

void format_node_defaults(int n)
{
    static int	first_time	= TRUE;
    NODEATTR	*na;
    int		first, i, from;

    if(n == DEFAULT) {
	na = &DEFAULTNODE;		first = 0;
	node_defaults[0].values[0]	= na->messagerate;
	node_defaults[1].values[0]	= na->minmessagesize;
	node_defaults[2].values[0]	= na->maxmessagesize;

	node_defaults[0].panel_value	=
	node_defaults[1].panel_value	=
	node_defaults[2].panel_value	= 0;
    }
    else {
	na = &(NP[n].nattr);		first = 1;
	for(i=0 ; i<N_NODE_DEFAULTS ; i++)
	    node_defaults[i].labels[0]	= strdup(" Default ");
	node_defaults[0].values[1]	=
		    WHICH(na->messagerate,    node_defaults[0].orig[1]);
	node_defaults[1].values[1]	=
		    WHICH(na->minmessagesize, node_defaults[1].orig[1]);
	node_defaults[2].values[1]	=
		    WHICH(na->maxmessagesize, node_defaults[2].orig[1]);

	node_defaults[0].panel_value = (na->messagerate == DEFAULT)    ? 0 : 1;
	node_defaults[1].panel_value = (na->minmessagesize == DEFAULT) ? 0 : 1;
	node_defaults[2].panel_value = (na->maxmessagesize == DEFAULT) ? 0 : 1;
    }
    for(from=0 ; from<N_NODE_DEFAULTS ; from++)
	for(i=(first+1) ; i<N_NODE_CHOICES ; i++)
	    node_defaults[from].values[i]	= node_defaults[from].orig[i];

    for(i=first ; i<N_NODE_CHOICES ; i++) {
	fmt_choice(node_defaults[0].values[i], 1000, "%ds","%dms");
	node_defaults[0].labels[i]	= strdup(chararray);
    }
    for(i=first ; i<N_NODE_CHOICES ; i++) {
	fmt_choice(node_defaults[1].values[i], K, "%dKB", "%d bytes");
	node_defaults[1].labels[i]	= strdup(chararray);
    }
    for(i=first ; i<N_NODE_CHOICES ; i++) {
	fmt_choice(node_defaults[2].values[i], K, "%dKB", "%d bytes");
	node_defaults[2].labels[i]	= strdup(chararray);
    }

    if(first_time)  {
	static int	default_displayed = FALSE;	/* never used */
	int		c, d, nn;
	char		buf[16];

	for(d=0 ; d<N_NODE_DEFAULTS ; ++d) {
	    sprintf(chararray,"node_choice_title(%d)",d);
	    Tcl_SetVar(tcl_interp, chararray,
			    node_defaults[d].title, TCL_GLOBAL_ONLY);
	}

	for(nn=DEFAULT ; nn<=0 ; ++nn)	/* nn just takes on DEFAULT and 0 */
	    for(d=0 ; d<N_NODE_DEFAULTS ; ++d) {
		sprintf(chararray,"node_choice_set(%d,%d)",nn,d);
		sprintf(buf      ,"%d", node_defaults[d].panel_value);
		Tcl_SetVar(tcl_interp, chararray, buf, TCL_GLOBAL_ONLY);

		for(c=0 ; c<N_NODE_CHOICES ; ++c) {
		    sprintf(chararray,"node_choice_label(%d,%d,%d)",nn,d,c);
		    Tcl_SetVar(tcl_interp, chararray,
				node_defaults[d].labels[c], TCL_GLOBAL_ONLY);
		}
	    }

	TCLTK_createcommand("node_debug_button", node_debug_button);

	strcpy(chararray, "stdio_quiet(default)");
	Tcl_LinkVar(tcl_interp, chararray,
			    (char *)&DEFAULTNODE.stdio_quiet, TCL_LINK_BOOLEAN);
	strcpy(chararray, "trace_events(default)");
	Tcl_LinkVar(tcl_interp, chararray,
			    (char *)&DEFAULTNODE.trace_all, TCL_LINK_BOOLEAN);

	strcpy(chararray, "node_displayed(default)");
	Tcl_LinkVar(tcl_interp, chararray,
			    (char *)&default_displayed, TCL_LINK_BOOLEAN);

	TCLTK("InitNode default default default .%s 0 0", argv0);
	first_time = FALSE;
    }
}

void init_nodewindow(int n)
{
    NODE	*np	= &NP[n];

    sprintf(chararray, "stdio_quiet(%d)", n);
    Tcl_LinkVar(tcl_interp, chararray,
		    (char *)&np->nattr.stdio_quiet, TCL_LINK_BOOLEAN);
    sprintf(chararray, "trace_events(%d)", n);
    Tcl_LinkVar(tcl_interp, chararray,
		    (char *)&np->nattr.trace_all, TCL_LINK_BOOLEAN);

    sprintf(chararray, "node_displayed(%d)", n);
    Tcl_LinkVar(tcl_interp, chararray,
			(char *)&np->displayed, TCL_LINK_BOOLEAN);

    format_node_defaults(n);

    TCLTK("InitNode %d %s %s .%s %d %d",
			    n, NP[n].nodename,
			    np->nodetype == NT_HOST ? "host" : "router", 
			    argv0,
			    np->nattr.winx, np->nattr.winy);
}

void toggle_node_window(int n)
{
    TCLTK("ToggleNodeAttributes %d", n);	/* never default */
}

/* ---------------------------------------------------------------------- */

int node_choice(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    NODECHOICE	nc;
    NODEATTR	*na;
    int		n, now = 0;

    if(vflag) {
	for(n=0 ; n<argc-1 ; ++n)
	    fprintf(stderr,"%s ", argv[n]);
	fprintf(stderr,"%s\n", argv[argc-1]);
    }
    if(argc != 4) {
	interp->result	= "wrong # args";
	return TCL_ERROR;
    }
    nc.n	= atoi(argv[1]);
    nc.r	= atoi(argv[2]);
    nc.c	= atoi(argv[3]);
    if(nc.n < -1 || nc.n >= _NNODES || nc.c < 0 || nc.c >= N_NODE_CHOICES) {
	interp->result	= "invalid node or choice #";
	return TCL_ERROR;
    }
    nc.value   =
        (nc.c == 0 && nc.n != DEFAULT) ? DEFAULT
                                       : node_defaults[nc.r].values[nc.c];

    na	= (nc.n==DEFAULT) ? &DEFAULTNODE : (&NP[nc.n].nattr);

    switch (nc.r) {
    case RATE_CHOICE:
	na->messagerate		= nc.value;
	now			= WHICH(nc.value, DEFAULTNODE.messagerate);
	break;
    case MINSIZE_CHOICE:
	na->minmessagesize	= nc.value;
	now			= WHICH(nc.value, DEFAULTNODE.minmessagesize);
	break;
    case MAXSIZE_CHOICE:
	na->maxmessagesize	= nc.value;
	now			= WHICH(nc.value, DEFAULTNODE.maxmessagesize);
	break;
    }
    if(dflag)
	fprintf(stderr,"%s.%s = %d\n",
		    (nc.n == DEFAULT) ? "defaultnode" : NP[nc.n].nodename,
		    node_defaults[nc.r].keyword, now);

    return TCL_OK;
}
#endif	/* defined(USE_TCLTK) */
