

typedef enum {
    SE_NULL = 0,
    SE_APPLICATIONREADY,	SE_TIMER,
    SE_NODESTATE,		SE_LINKSTATE,
    SE_REPORT,			SE_TIMESUP,
#if	defined(USE_TCLTK)
    SE_DEBUGBUTTON,		SE_KEYBOARDREADY,
    SE_DRAWFRAME,		SE_PINGX
#endif
} SCHEDULE;

typedef struct _sq {
    struct _sq		*next;
    SCHEDULE		se;
    int			node;
    CnetEvent		ne;
    long		ms;
    CnetTimer		timer;
    CnetData		data;
} SCHEDULE_Q;


static	SCHEDULE_Q	GSQ	= { (SCHEDULE_Q *)NULL };
static	int		nqueued	= 0;


/* --------------------------------------------------------------------- */


/*  We keep the Global Schedule Queue sorted by:

	time	- so that events occur in order,
	node	- to minimize context switches,
	ne	- so that ack frames arrive before their timeouts
 */

static void insert_event(SCHEDULE se, CnetEvent	ne, int	node, long abs_ms,
			 CnetTimer timer, CnetData data)
{
    SCHEDULE_Q	*s	= &GSQ;
    SCHEDULE_Q	*t;

    t		= NEW(SCHEDULE_Q);
    t->se	= se;
    t->node	= node;
    t->ne	= ne;
    t->ms	= abs_ms;
    t->timer	= timer;
    t->data	= data;

    while(s->next) {
	if(abs_ms  < s->next->ms)
	    break;
	if(abs_ms == s->next->ms) {
	    if(node  < s->next->node)
		break;
	    if(node == s->next->node && (int)ne < (int)s->next->ne)
		break;
	}
	s	= s->next;
    }
    t->next	= s->next;
    s->next	= t;
    ++nqueued;
}

void schedule_event(CnetEvent ne, int node, long rel_ms,
		    CnetTimer timer, CnetData data)
{
    SCHEDULE	se;

    if(node < 0) {		/* this event is for a link, not a node */
	se	= SE_LINKSTATE;
	node	= -node-1;
    }
    else if(IS_TIMER(ne))
	se	= SE_TIMER;
#if	defined(USE_TCLTK)
    else if(ne == EV_DRAWFRAME)
	se	= SE_DRAWFRAME;
    else if(IS_DEBUG(ne))
	se	= SE_DEBUGBUTTON;
    else if(ne == EV_KEYBOARDREADY)
	se	= SE_KEYBOARDREADY;
#endif
    else if(ne == EV_REBOOT)
	se	= SE_NODESTATE;
    else
	return;

    insert_event(se,ne, node, TIMENOW_in_MSEC + (rel_ms==0?1:rel_ms),
		    timer, data);
}

int unschedule_timer(CnetTimer timer, CnetData *datawas)
{					/* reqd by: CNET_stop_timer() */
    SCHEDULE_Q	*t;
    SCHEDULE_Q	*s	= &GSQ;

    while((t = s->next)) {
	if(timer == t->timer && THISNODE == t->node && IS_TIMER(t->ne)) {
	    if(datawas) {
		*datawas	= t->data;
		return(TRUE);
	    }
	    s->next	= t->next;
	    free((char *)t);
	    --nqueued;
	    return(TRUE);
	}
	s	= s->next;
    }
    return(FALSE);
}

void filter_node_events(int n)	/* reqd by: node_menu.c */
{
    SCHEDULE_Q	*s	= &GSQ;
    SCHEDULE_Q	*t;

    while((t = s->next)) {
	if(t->node == n && t->se != SE_LINKSTATE) {
	    s->next	= t->next;
	    free((char *)t);
	    --nqueued;
	}
	s	= s->next;
    }
}
