#include "cnetheader.h"
#include "statistics.h"

/*  The cnet network simulator (v1.7.7)
    Copyright (C) 1992-2002, Chris McDonald

    Chris McDonald, chris@cs.uwa.edu.au
    Department of Computer Science & Software Engineering
    The University of Western Australia,
    Crawley, Western Australia, 6009
    PH: +61 8 9380 2533, FAX: +61 8 9380 1089.

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

static char *stats_titles[] = {
    "Simulation time",
    "Events raised",
    "Messages generated",
    "Messages delivered",
    "Errors detected",
    "Efficiency (bytes AL/PL)",
    "Average delivery rate",
    "Average delivery time",
    "Frames transmitted",
    "Frames received",
    "Frames corrupted",
    "Frames lost",
    "Total transmission cost",
    "ms",
    "%",
    "KB/s",
    "ms"
};

#define	N_STATS			13
#define	N_STATS_TITLES		(sizeof(stats_titles)/sizeof(stats_titles[0]))

#define	P(n)			n, ((n)==1 ? "" : "S")


/* ------------------ Global stats information first ------------------ */


static	int	STATS[N_STATS]		= {0};
static	long	frame_txbytes		= 0L;
static	long	msg_rxbytes		= 0L;
static	int	sflag			= FALSE;
static	int	zflag			= FALSE;


void init_stats_layer(int s, int z)
{
    sflag			= s;
    zflag			= z;
    STATS[(int)STATS_MSGRATE]	= 1;
}

void stats_summary()
{
    int		i, r, s;

    if(!sflag)
	return;

    for(i=0, r=0 ; i<_NNODES ; ++i)
	if(NP[i].nodetype == NT_ROUTER) r++;

    fprintf(stdout, "%d HOST%s, %d ROUTER%s and %d LINK%s\n",
			P(_NNODES-r), P(r), P(_NLINKS) );

    STATS[(int)STATS_TIME]	= TIMENOW_in_MSEC;
    for(i=0 ; i<N_STATS ; ++i)
	if((s=STATS[i]) != 0 || zflag) {

		if(i == (int)STATS_MSGRATE) {
		    double rate = 0.0;

		    if(msg_rxbytes != 0)
			rate = (msg_rxbytes * 1.024) / (float)TIMENOW_in_MSEC;
		    fprintf(stdout, "%-24s: %10.2f\n",
			    stats_titles[(int)STATS_MSGRATE], rate);
		}
		else if(i == (int)STATS_MSGEFFICIENCY) {
		    double eff = 0.0;

		    if(frame_txbytes != 0)
			eff = (msg_rxbytes / (float)frame_txbytes) * 100.0;
		    fprintf(stdout, "%-24s: %10.1f\n",
			    stats_titles[(int)STATS_MSGEFFICIENCY], eff);
		}
		else {
		    if(i == (int)STATS_TIMEDELIVERY && s != 0)
			s = STATS[(int)STATS_TIMEDELIVERY] /
			    STATS[(int)STATS_MSGDELIVERED];
		    fprintf(stdout,"%-24s: %10d\n",stats_titles[i],s);
		}
	}

    for(i=0 ; i<N_CNET_EVENTS ; ++i)
	if(EV_COUNT[i] != 0 || zflag)
	    fprintf(stdout,"%-24s: %10ld\n", cnet_evname[i],EV_COUNT[i]);
    fflush(stdout);
}


void inc_stats(STATS_TYPE type, int delta1, int delta2)
{
    STATS[(int)type]++;

    if(type == STATS_FRAMESTX) {
	STATS[(int)STATS_FRAMECOST]	+= delta1;
	frame_txbytes			+= delta2;
	STATS[(int)STATS_MSGEFFICIENCY]	+= 1;
    }
    else if(type == STATS_MSGDELIVERED) {
	STATS[(int)STATS_TIMEDELIVERY]	+= delta1;
	msg_rxbytes			+= delta2;
    }
}


/* ----------------- only windowing-specific stats follows ---------------- */

#if	defined(USE_TCLTK)

static	int	stats_displayed		= FALSE;
static	long	msgs_ok			= 0L;
static	long	msgs_err		= 0L;

void init_statswindow()
{
    int	n;

    for(n=0 ; n<N_STATS_TITLES ; ++n) {
	sprintf(chararray, "STATS_TITLE(%d)", n);
	Tcl_SetVar(tcl_interp,chararray,stats_titles[n],TCL_GLOBAL_ONLY);
    }
    Tcl_SetVar(tcl_interp,"UPDATE_TITLE",UPDATE_TITLE,TCL_GLOBAL_ONLY);
    Tcl_LinkVar(tcl_interp, "stats_displayed",
				(char *)&stats_displayed, TCL_LINK_BOOLEAN);
}

void flush_stats()
{
    extern void flush_load(int, int);

    static int	last_s	= 0;
    int		s;

    STATS[(int)STATS_TIME]	= TIMENOW_in_MSEC;
    if(stats_displayed) {

	int	i;
	char	cmdbuf[BUFSIZ];
	char	*p;

	strcpy(cmdbuf, "UpdateStatistics");
	p	= cmdbuf;
	while(*p)
	    ++p;
	*p++	= ' ';
	*p++	= '{';
	for(i=0 ; i<N_STATS ; ++i) {
	    *p++	= ' ';
	    *p++	= '"';
	    if(i == (int)STATS_MSGRATE)
		sprintf(p,"%10.2f", (msg_rxbytes*1.024)/(float)TIMENOW_in_MSEC);
	    else if(i == (int)STATS_MSGEFFICIENCY) {
		if(frame_txbytes == 0)
		    sprintf(p,"%10c", '-');
		else
		    sprintf(p,"%10.1f", (msg_rxbytes/(float)frame_txbytes)*100);
	    }
	    else if(i == (int)STATS_TIMEDELIVERY) {
		if(STATS[(int)STATS_MSGDELIVERED] == 0)
		     sprintf(p,"%10c", '-');
		else sprintf(p,"%10d", 
		STATS[(int)STATS_TIMEDELIVERY]/STATS[(int)STATS_MSGDELIVERED]);
	    }
	    else if(STATS[i] == 0)
		 sprintf(p,"%10c", '-');
	    else sprintf(p,"%10d", STATS[i]);

	    p		+= 10;
	    *p++	 = '"';
	}
	*p++	= '}';
	*p	= '\0';
	TCLTK(cmdbuf);
    }
    flush_load(STATS[(int)STATS_MSGDELIVERED], STATS[(int)STATS_FRAMESTX]);

    s	= msgs_ok + msgs_err;
    if(s != last_s) {
	sprintf(chararray,"%6ld (%5.1f%%)", msgs_ok,(float)msgs_ok/s*100.0);
	Tcl_SetVar(tcl_interp, "CN_DELIVERY", chararray, TCL_GLOBAL_ONLY);
	last_s	= s;
    }
}

/* --------------- Per-node TCL/TK stats code follows ---------------- */

void inc_nodestats(int which_stat, int bytes)
{
    NODE	*np;
    int		c;

/*  which_stat	= 0	=> GENERATED,
		= 1	=> RECEIVED CORRECTLY,
		= 2	=> RECEIVED IN ERROR
 */
         if(which_stat == 1)	++msgs_ok;
    else if(which_stat == 2)	++msgs_err;

    c	= which_stat*2;

    np	= &NP[THISNODE];
    np->stats_count[c]		+= 1;
    np->stats_count[c+1]	+= bytes;
}

void flush_nodestats()
{
    NODE	*np;
    int		n;

    for(n=0, np=NP ; n<_NNODES ; ++n, ++np)
	if(np->displayed && np->nodetype == NT_HOST)
	    TCLTK(
"UpdateNodeStats %d {\"%10ld\" \"%10ld\" \"%10.2f\" \"%10ld\" \"%10ld\" \"%10.2f\" \"%10ld\" \"%10ld\"}",
		n,
		np->stats_count[0], np->stats_count[1],
			    ((np->stats_count[1] * 1.024) / TIMENOW_in_MSEC),
		np->stats_count[2], np->stats_count[3],
			    ((np->stats_count[3] * 1.024) / TIMENOW_in_MSEC),
		np->stats_count[4], np->stats_count[5]);
}

/* ------------------ Per-link TCL/TK stats code follows ------------- */

void inc_linkstats(int link, int which_stat, int bytes)
{
    int		c, which_link;
    LINKPANEL	*lp;

/*  which_stat	= 0	=> TRANSMITTED,
		= 1	=> RECEIVED,
		= 2	=> ERRORS INTRODUCED
 */

    which_link	= NP[THISNODE].links[link-1];
    if(LP[which_link].endmin == THISNODE)
	lp      = &(LP[which_link].panelmin);
    else
	lp      = &(LP[which_link].panelmax);

    if(which_stat == 0)
	LP[which_link].ntxed += bytes;
    c		= which_stat*2;

    lp->count[c]	+= 1;
    lp->count[c+1]	+= bytes;
}

static void flush_1_linkstat(LINKPANEL *lp, int from, int to)
{
    TCLTK(
"UpdateLinkStats %d %d {\"%10ld\" \"%10ld\" \"%10.2f\" \"%10ld\" \"%10ld\" \"%10.2f\" \"%10ld\" \"%10ld\"}",
	from, to,
	lp->count[0], lp->count[1], (lp->count[1] * 1.024) / TIMENOW_in_MSEC,
	lp->count[2], lp->count[3], (lp->count[3] * 1.024) / TIMENOW_in_MSEC,
	lp->count[4], lp->count[5]);
}

void flush_linkstats()
{
    int		l;
    LINKPANEL	*lp;

    for(l=0 ; l<_NLINKS ; ++l) {
	lp	= &(LP[l].panelmin);
	if(lp->displayed)
	    flush_1_linkstat(lp, LP[l].endmin, LP[l].endmax);

	lp	= &(LP[l].panelmax);
	if(lp->displayed)
	    flush_1_linkstat(lp, LP[l].endmax, LP[l].endmin);
    }
}

#endif	/* defined(USE_TCLTK) */
